/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.swing.ImageIcon;

import org.apache.log4j.Logger;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.config.AttachmentConfig;
import com.eviware.soapui.config.CallConfig;
import com.eviware.soapui.config.CredentialsConfig;
import com.eviware.soapui.impl.wsdl.submit.RequestTransportRegistry;
import com.eviware.soapui.impl.wsdl.submit.filters.PropertyExpansionRequestFilter;
import com.eviware.soapui.impl.wsdl.submit.transports.http.AttachmentUtils;
import com.eviware.soapui.impl.wsdl.submit.transports.http.WsdlResponse;
import com.eviware.soapui.impl.wsdl.support.CompressedStringSupport;
import com.eviware.soapui.impl.wsdl.support.FileAttachment;
import com.eviware.soapui.impl.wsdl.support.ModelItemIconAnimator;
import com.eviware.soapui.impl.wsdl.support.RequestFileAttachment;
import com.eviware.soapui.model.iface.Attachment;
import com.eviware.soapui.model.iface.Interface;
import com.eviware.soapui.model.iface.MessagePart;
import com.eviware.soapui.model.iface.Request;
import com.eviware.soapui.model.iface.Submit;
import com.eviware.soapui.model.iface.SubmitContext;
import com.eviware.soapui.model.iface.SubmitListener;
import com.eviware.soapui.model.support.InterfaceListenerAdapter;
import com.eviware.soapui.settings.WsdlSettings;
import com.eviware.soapui.support.UISupport;
import com.eviware.soapui.support.types.StringToStringMap;

/**
 * Request implementation holding a SOAP request
 * 
 * @author Ole.Matzura
 */

public class WsdlRequest extends AbstractWsdlModelItem<CallConfig> implements Request, AttachmentContainer
{
	public final static Logger log = Logger.getLogger( WsdlRequest.class );
	
	public static final String RESPONSE_PROPERTY = WsdlRequest.class.getName() + "@response";
	public static final String RESPONSE_CONTENT_PROPERTY = WsdlRequest.class.getName() + "@response-content";
	public static final String ATTACHMENTS_PROPERTY = WsdlRequest.class.getName() + "@attachments";
	public static final String INLINE_RESPONSE_ATTACHMENTS = WsdlRequest.class.getName() + "@inline-response-attachments";
	public static final String EXPAND_MTOM_RESPONSE_ATTACHMENTS = WsdlRequest.class.getName() + "@expand-mtom-attachments";
	public static final String STRIP_WHITESPACES = WsdlRequest.class.getName() + "@strip-whitespaces";
	public static final String REMOVE_EMPTY_CONTENT = WsdlRequest.class.getName() + "@remove_empty_content";
	public static final String REQUEST_HEADERS_PROPERTY = WsdlRequest.class.getName() + "@request-headers";
	public static final String ENCODE_ATTACHMENTS = WsdlRequest.class.getName() + "@encode_attachments";
	public static final String DISABLE_MULTIPART_ATTACHMENTS = WsdlRequest.class.getName() + "@disable-multipart-attachments";
	public static final String WSS_TIMETOLIVE = WsdlRequest.class.getName() + "@wss-time-to-live";
	public static final String BIND_ADDRESS = WsdlRequest.class.getName() + "@bind_address";
	public static final String OPERATION_PROPERTY = WsdlRequest.class.getName() + "@operation";
	
	public final static String PW_TYPE_NONE="None";
   public final static String PW_TYPE_DIGEST="PasswordDigest";
   public final static String PW_TYPE_TEXT="PasswordText";
   
   private WsdlOperation operation;
   private WsdlResponse response;
   protected List<FileAttachment> attachments = new ArrayList<FileAttachment>();
   
	private RequestIconAnimator iconAnimator;
   private Set<SubmitListener> listeners = new HashSet<SubmitListener>();
	private List<WsdlAttachmentPart> definedAttachmentParts;
	private InternalInterfaceListener interfaceListener = new InternalInterfaceListener();
	private String requestContent;

   public WsdlRequest( WsdlOperation operation, CallConfig callConfig )
   {
   	this( operation, callConfig, false );
   }
	
   public WsdlRequest( WsdlOperation operation, CallConfig callConfig, boolean forLoadTest )
   {
   	super( callConfig, operation, null );
   	
      this.operation = operation;
     
      initEndpoints();
      initAttachments();
      
      // ensure encoding
      if( callConfig.getEncoding() == null || callConfig.getEncoding().length() == 0 )
      {
      	callConfig.setEncoding( "UTF-8" );
      }

      if( !forLoadTest )
      {
	      iconAnimator = initIconAnimator();
	     	addSubmitListener( iconAnimator );
	    
	      operation.getInterface().addPropertyChangeListener( interfaceListener );
	      operation.getInterface().addInterfaceListener( interfaceListener );
      }
   }
   
   private void initAttachments()
	{
		for( AttachmentConfig ac : getConfig().getAttachmentList() )
		{
			FileAttachment attachment = new RequestFileAttachment( ac, this );
			attachments.add( attachment);
		}
	}

	public void updateConfig(CallConfig request)
	{
		setConfig( request );
	}
   
   public ModelItemIconAnimator getIconAnimator()
   {
   	return iconAnimator;
   }

   protected RequestIconAnimator initIconAnimator()
   {
   	return new RequestIconAnimator();
   }

   protected void initEndpoints()
   {
   	if( getEndpoint() == null )
   	{
	      String[] endpoints = operation.getInterface().getEndpoints();
	      if( endpoints.length > 0 )
	      {
	         setEndpoint( endpoints[0] );
	      }
   	}
   }

   public String getRequestContent()
   {
   	if( getConfig().getRequest() == null )
   		getConfig().addNewRequest();
   	
   	if( requestContent == null )
   		requestContent = CompressedStringSupport.getString( getConfig().getRequest() );
   	
   	return requestContent;
   }
   
   public void setEndpoint(String endpoint)
   {
      String old = getEndpoint();
      if( old != null && old.equals( endpoint ))
      	return;
      
      getConfig().setEndpoint( endpoint );
      notifyPropertyChanged( ENDPOINT_PROPERTY, old, endpoint);
   }

   public String getEndpoint()
   {
      return getConfig().getEndpoint();
   }

   public String getEncoding()
   {
      return getConfig().getEncoding();
   }

   public void setEncoding(String encoding)
   {
      String old = getEncoding();
      getConfig().setEncoding( encoding );
      notifyPropertyChanged( ENCODING_PROPERTY, old, encoding );
   }
   
   public StringToStringMap getRequestHeaders()
   {
   	return StringToStringMap.fromXml( getSettings().getString( REQUEST_HEADERS_PROPERTY, null ));
   }
   
   public void setRequestHeaders( StringToStringMap map )
   {
   	StringToStringMap old = getRequestHeaders();
   	getSettings().setString( REQUEST_HEADERS_PROPERTY, map.toXml() );
   	notifyPropertyChanged( REQUEST_HEADERS_PROPERTY, old, map );
   }
   
   public boolean isInlineResponseAttachments()
   {
   	return getSettings().getBoolean( INLINE_RESPONSE_ATTACHMENTS );
   }
   
   public void setInlineResponseAttachments( boolean inlineResponseAttachments )
   {
   	boolean old = getSettings().getBoolean( INLINE_RESPONSE_ATTACHMENTS );
   	getSettings().setBoolean( INLINE_RESPONSE_ATTACHMENTS, inlineResponseAttachments );
   	notifyPropertyChanged( INLINE_RESPONSE_ATTACHMENTS, old, inlineResponseAttachments );
   }
   
   public boolean isStripWhitespaces()
   {
   	return getSettings().getBoolean( STRIP_WHITESPACES );
   }
   
   public void setStripWhitespaces( boolean stripWhitespaces )
   {
   	boolean old = getSettings().getBoolean( STRIP_WHITESPACES );
   	getSettings().setBoolean( STRIP_WHITESPACES, stripWhitespaces );
   	notifyPropertyChanged( STRIP_WHITESPACES, old, stripWhitespaces );
   }
   
   public boolean isExpandMtomResponseAttachments()
   {
   	return getSettings().getBoolean( EXPAND_MTOM_RESPONSE_ATTACHMENTS );
   }
   
   public void setExpandMtomResponseAttachments( boolean expandMtomResponseAttachments )
   {
   	boolean old = getSettings().getBoolean( EXPAND_MTOM_RESPONSE_ATTACHMENTS );
   	getSettings().setBoolean( EXPAND_MTOM_RESPONSE_ATTACHMENTS, expandMtomResponseAttachments );
   	notifyPropertyChanged( EXPAND_MTOM_RESPONSE_ATTACHMENTS, old, expandMtomResponseAttachments );
   }
   
   /**
    * Use getResponse().getContentAsString();
    * @deprecated
    */
   
   public String getResponseContent()
   {
      return response == null ? null : response.getContentAsString();
   }

   public WsdlResponse getResponse()
   {
   	return response;
   }
   
   public WsdlOperation getOperation()
   {
      return operation;
   }
   
   public void setOperation( WsdlOperation wsdlOperation )
	{
   	WsdlOperation oldOperation = operation;
		this.operation = wsdlOperation;
		
		definedAttachmentParts = null;
		notifyPropertyChanged( OPERATION_PROPERTY, oldOperation, operation );
	}

   public void setRequestContent(String request)
   {
      String old = getRequestContent();
      if( request.equals( old ))
      	return;
      
      requestContent = request;
      definedAttachmentParts = null;
      notifyPropertyChanged( REQUEST_PROPERTY, old, request );
   }
   
   public void setResponse( WsdlResponse response, SubmitContext context )
   {
   	WsdlResponse oldResponse = (WsdlResponse) getResponse();
		this.response = response;
		
      notifyPropertyChanged( RESPONSE_PROPERTY, oldResponse, response );
   }

   public ImageIcon getIcon()
   {
   	return iconAnimator.getIcon();
   }

   public String getUsername()
   {
   	CredentialsConfig credentialsConfig = getConfig().getCredentials();
   	if( credentialsConfig == null ) return null;
   	
   	return credentialsConfig.getUsername();
   }
   
   public String getPassword()
   {
   	CredentialsConfig credentialsConfig = getConfig().getCredentials();
   	if( credentialsConfig == null ) return null;
   	
   	return credentialsConfig.getPassword();
   }
   
   public String getDomain()
   {
   	CredentialsConfig credentialsConfig = getConfig().getCredentials();
   	if( credentialsConfig == null ) return null;
   	
   	return credentialsConfig.getDomain();
   }
   
   public void setUsername( String username )
   {
   	CredentialsConfig credentialsConfig = getConfig().getCredentials();
   	if( credentialsConfig == null ) 
   		credentialsConfig = getConfig().addNewCredentials();
   	
   	credentialsConfig.setUsername( username );
   }
   
   public void setPassword( String password )
   {
   	CredentialsConfig credentialsConfig = getConfig().getCredentials();
   	if( credentialsConfig == null ) 
   		credentialsConfig = getConfig().addNewCredentials();
   	
   	credentialsConfig.setPassword( password );
   }
   
   public void setDomain( String domain )
   {
   	CredentialsConfig credentialsConfig = getConfig().getCredentials();
   	if( credentialsConfig == null ) 
   		credentialsConfig = getConfig().addNewCredentials();
   	
   	credentialsConfig.setDomain( domain );
   }
   
   public void addSubmitListener(SubmitListener listener) 
	{
		listeners.add( listener );
	}

	public void removeSubmitListener(SubmitListener listener) 
	{
		listeners.remove( listener );
	}

	public WsdlSubmit submit( SubmitContext submitContext, boolean async ) throws SubmitException
	{
      String endpoint = PropertyExpansionRequestFilter.expandProperties( submitContext, getEndpoint());
		if( endpoint == null || endpoint.trim().length() == 0 )
      {
      	UISupport.showErrorMessage( "Missing endpoint for request [" + getName() + "]" );
      	return null;
      }
		
		try
		{
			WsdlSubmit submitter = new WsdlSubmit(this, (SubmitListener[]) listeners.toArray(new SubmitListener[listeners
					.size()]), RequestTransportRegistry.getTransport(endpoint, submitContext));
			submitter.submitRequest(submitContext, async);
			return submitter;
		}
		catch( Exception e )
		{
			throw new SubmitException( e.toString() );
		}
	}
	
	private class InternalInterfaceListener extends InterfaceListenerAdapter implements PropertyChangeListener
	{
		public void propertyChange(PropertyChangeEvent evt)
		{
			if( evt.getPropertyName().equals( Interface.ENDPOINT_PROPERTY ))
			{
				String endpoint = getEndpoint();
				if( evt.getOldValue() != null && evt.getOldValue().equals( endpoint ))
				{
					setEndpoint( (String) evt.getNewValue() );
				}
			}
		}
	}

	public String getWssPasswordType()
	{
		return getConfig().getWssPasswordType();
	}

	public void setWssPasswordType(String wssPasswordType)
	{
		getConfig().setWssPasswordType( wssPasswordType );
	}

	/* (non-Javadoc)
	 * @see com.eviware.soapui.impl.wsdl.AttachmentContainer#attachFile(java.io.File, boolean)
	 */
	public Attachment attachFile( File file, boolean cache )
	{
		try
		{
			FileAttachment fileAttachment = new RequestFileAttachment( file, cache, this );
			attachments.add( fileAttachment );
			notifyPropertyChanged(ATTACHMENTS_PROPERTY, null, fileAttachment );
			return fileAttachment;
		}
		catch (IOException e)
		{
			UISupport.showErrorMessage( e );
			return null;
		}
	}
	
	/* (non-Javadoc)
	 * @see com.eviware.soapui.impl.wsdl.AttachmentContainer#getAttachmentCount()
	 */
	public int getAttachmentCount()
	{
		return attachments.size();
	}
	
	/* (non-Javadoc)
	 * @see com.eviware.soapui.impl.wsdl.AttachmentContainer#getAttachmentAt(int)
	 */
	public Attachment getAttachmentAt( int index )
	{
		return attachments.get( index );
	}
	
	/* (non-Javadoc)
	 * @see com.eviware.soapui.impl.wsdl.AttachmentContainer#getAttachmentsForPart(java.lang.String)
	 */
	public Attachment [] getAttachmentsForPart( String partName )
	{
		List<Attachment> result = new ArrayList<Attachment>();
		
		for( Attachment attachment : attachments )
		{
			if( attachment.getPart().equals( partName ))
				result.add( attachment );
		}
		
		return result.toArray( new Attachment[result.size()]);
	}
	
	/* (non-Javadoc)
	 * @see com.eviware.soapui.impl.wsdl.AttachmentContainer#removeAttachment(com.eviware.soapui.model.iface.Attachment)
	 */
	public void removeAttachment( Attachment attachment )
	{
		int ix = attachments.indexOf( attachment );
		attachments.remove( ix );
		
		try
		{
			notifyPropertyChanged(ATTACHMENTS_PROPERTY, attachment, null );
		}
		finally
		{
			getConfig().removeAttachment( ix);
		}
	}
	
	/* (non-Javadoc)
	 * @see com.eviware.soapui.impl.wsdl.AttachmentContainer#getAttachments()
	 */
	public Attachment[] getAttachments()
	{
		return attachments.toArray( new Attachment[attachments.size()] );
	}
	
	/* (non-Javadoc)
	 * @see com.eviware.soapui.impl.wsdl.AttachmentContainer#getDefinedAttachmentParts()
	 */
	public WsdlAttachmentPart [] getDefinedAttachmentParts()
	{
		if( definedAttachmentParts == null )
		{
			try
			{
				UISupport.setHourglassCursor();
				definedAttachmentParts = AttachmentUtils.extractAttachmentParts( 
							operation, getRequestContent(), true, false );
			}
			catch (Exception e)
			{
				log.warn( e.toString() );
				definedAttachmentParts = new ArrayList<WsdlAttachmentPart>();
			}
			finally 
			{
				UISupport.resetCursor();
			}
		}
		
		return definedAttachmentParts.toArray( new WsdlAttachmentPart[definedAttachmentParts.size()] );
	}
	
	/* (non-Javadoc)
	 * @see com.eviware.soapui.impl.wsdl.AttachmentContainer#getAttachmentPart(java.lang.String)
	 */
	public WsdlAttachmentPart getAttachmentPart( String partName )
	{
		WsdlAttachmentPart[] parts = getDefinedAttachmentParts();
		for( WsdlAttachmentPart part : parts )
		{
			if( part.getName().equals( partName ))
				return part;
		}
		
		return null;
	}

	public void copyAttachmentsTo(WsdlRequest newRequest)
	{
		if( getAttachmentCount() > 0 )
      {
			try
			{
				UISupport.setHourglassCursor();
				for (int c = 0; c < getAttachmentCount(); c++)
				{
					try
					{
						Attachment attachment = getAttachmentAt(c);
						newRequest.addAttachment( attachment );
					}
					catch (Exception e)
					{
						SoapUI.logError( e );
					}					
				}
			}
			finally 
			{
				UISupport.resetCursor();
			}      	
      }
	}

	private Attachment addAttachment(Attachment attachment)
	{
		if( attachment instanceof FileAttachment )
		{
			AttachmentConfig oldConfig = ((FileAttachment)attachment).getConfig();
			AttachmentConfig newConfig = (AttachmentConfig) getConfig().addNewAttachment().set( oldConfig );
			FileAttachment newAttachment = new RequestFileAttachment( newConfig, this );
			attachments.add( newAttachment);
			return newAttachment;
		}
		else log.error( "Unkown attachment type: " + attachment );
		
		return null;
	}

	public void copyTo(WsdlRequest newRequest, boolean copyAttachments, boolean copyHeaders)
	{
      newRequest.setEncoding( getEncoding() );
      newRequest.setEndpoint( getEndpoint() );
      newRequest.setRequestContent( getRequestContent() );
      newRequest.setWssPasswordType( getWssPasswordType() );
      
      CredentialsConfig credentials = getConfig().getCredentials();
      if( credentials != null)
      	newRequest.getConfig().setCredentials( (CredentialsConfig) credentials.copy() );

      if( copyAttachments )
      	copyAttachmentsTo( newRequest );
      
      if( copyHeaders )
      	newRequest.setRequestHeaders( getRequestHeaders() );
	}
	
	/* (non-Javadoc)
	 * @see com.eviware.soapui.impl.wsdl.AttachmentContainer#isMtomEnabled()
	 */
	public boolean isMtomEnabled()
	{
		return getSettings().getBoolean( WsdlSettings.ENABLE_MTOM );
	}
	
	public void setMtomEnabled( boolean mtomEnabled )
	{
		getSettings().setBoolean( WsdlSettings.ENABLE_MTOM, mtomEnabled );
		definedAttachmentParts = null;
	}

	public void release()
	{
		super.release();
		
		getOperation().getInterface().removeInterfaceListener( interfaceListener );
		getOperation().getInterface().removePropertyChangeListener( interfaceListener );
	}

	public MessagePart[] getRequestParts()
	{
		try
		{
			List<MessagePart> result = new ArrayList<MessagePart>();
			result.addAll( Arrays.asList( getOperation().getDefaultRequestParts() ));
			result.addAll( Arrays.asList( getDefinedAttachmentParts()) );
			
			return result.toArray( new MessagePart[result.size()] );
		}
		catch (Exception e)
		{
			SoapUI.logError( e );
			return new MessagePart [0];
		}		
	}
	
	public MessagePart[] getResponseParts()
	{
		try
		{
			List<MessagePart> result = new ArrayList<MessagePart>();
			result.addAll( Arrays.asList( getOperation().getDefaultResponseParts() ));
			
			if( getResponse() != null )
				result.addAll( AttachmentUtils.extractAttachmentParts( 
						(WsdlOperation)getOperation(), getResponse().getContentAsString(), true, true ));
			
			return result.toArray( new MessagePart[result.size()] );
		}
		catch (Exception e)
		{
			SoapUI.logError( e );
			return new MessagePart [0];
		}		
	}
	
	protected class RequestIconAnimator extends ModelItemIconAnimator implements SubmitListener
	{
		public RequestIconAnimator()
		{
			super( WsdlRequest.this, "/request.gif", 
						new String[] {"/exec_request_1.gif", "/exec_request_2.gif",
						"/exec_request_3.gif", "/exec_request_4.gif"} );
		}
		
		public boolean beforeSubmit(Submit submit, SubmitContext context) 
		{
			if( isEnabled() && submit.getRequest() == getTarget() )
				start();
	      return true;
		}

		public void afterSubmit(Submit submit, SubmitContext context) 
		{
			if( submit.getRequest() == getTarget() )
				stop();
		}
	}

	public boolean isMultipartEnabled()
	{
		return !getSettings().getBoolean( DISABLE_MULTIPART_ATTACHMENTS );
	}
	
	public void setMultipartEnabled( boolean multipartEnabled )
	{
		getSettings().setBoolean( DISABLE_MULTIPART_ATTACHMENTS, !multipartEnabled );
	}

	public String getWssTimeToLive()
	{
		return getSettings().getString( WSS_TIMETOLIVE, null );
	}
	
	public void setWssTimeToLive( String ttl )
	{
		getSettings().setString( WSS_TIMETOLIVE, ttl );
	}

	@Override
	public void onSave()
	{
		if( requestContent != null )
		{
			CompressedStringSupport.setString( getConfig().getRequest(), requestContent );
			requestContent = null;
		}
	}
	
	public long getContentLength()
	{
		return getRequestContent().length();
	}

	public boolean isRemoveEmptyContent()
	{
		return getSettings().getBoolean( REMOVE_EMPTY_CONTENT );
	}
	
	public void setRemoveEmptyContent( boolean removeEmptyContent )
   {
   	boolean old = getSettings().getBoolean( REMOVE_EMPTY_CONTENT );
   	getSettings().setBoolean( REMOVE_EMPTY_CONTENT, removeEmptyContent );
   	notifyPropertyChanged( REMOVE_EMPTY_CONTENT, old, removeEmptyContent );
   }
	
	public boolean isEncodeAttachments()
	{
		return getSettings().getBoolean( ENCODE_ATTACHMENTS );
	}
	
	public void setEncodeAttachments( boolean encodeAttachments )
   {
   	boolean old = getSettings().getBoolean( ENCODE_ATTACHMENTS );
   	getSettings().setBoolean( ENCODE_ATTACHMENTS, encodeAttachments );
   	notifyPropertyChanged( ENCODE_ATTACHMENTS, old, encodeAttachments );
   }
	
	public String getBindAddress()
	{
		return getSettings().getString( BIND_ADDRESS, "" );
	}
	
	public void setBindAddress( String bindAddress )
   {
   	String old = getSettings().getString( BIND_ADDRESS, "" );
   	getSettings().setString( BIND_ADDRESS, bindAddress );
   	notifyPropertyChanged( BIND_ADDRESS, old, bindAddress );
   }
}
