/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.actions.iface;

import java.io.File;
import java.util.List;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.impl.wsdl.WsdlInterface;
import com.eviware.soapui.impl.wsdl.WsdlOperation;
import com.eviware.soapui.impl.wsdl.WsdlRequest;
import com.eviware.soapui.impl.wsdl.testcase.WsdlTestCase;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestRequest;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestRequestStep;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestStep;
import com.eviware.soapui.model.iface.Request;
import com.eviware.soapui.model.testsuite.TestCase;
import com.eviware.soapui.model.testsuite.TestSuite;
import com.eviware.soapui.support.UISupport;
import com.eviware.soapui.support.action.support.AbstractSoapUIAction;
import com.eviware.soapui.support.xml.XmlUtils;
import com.eviware.x.form.XFormDialog;
import com.eviware.x.form.XFormField;
import com.eviware.x.form.XFormFieldListener;
import com.eviware.x.form.support.ADialogBuilder;
import com.eviware.x.form.support.AField;
import com.eviware.x.form.support.AForm;
import com.eviware.x.form.support.AField.AFieldType;

/**
 * Updates the definition of a WsdlInterface.
 * 
 * @author Ole.Matzura
 */

public class UpdateInterfaceAction extends AbstractSoapUIAction<WsdlInterface>
{
   public static final String SOAPUI_ACTION_ID = "UpdateInterfaceAction";
   private XFormDialog dialog = null;

	public UpdateInterfaceAction()
   {
      super("Update Definition", "Reloads the definition for this interface and its operations");
   }

   public void perform( WsdlInterface iface, Object param )
	{
   	if( RemoveInterfaceAction.hasRunningDependingTests( iface ))
   	{
   		UISupport.showErrorMessage( "Cannot update Interface due to running depending tests" );
   		return;
   	}
   	
   	if( dialog == null )
   	{
   		dialog = ADialogBuilder.buildDialog( Form.class );
   		dialog.setBooleanValue( Form.CREATE_REQUESTS, true );
   		dialog.getFormField( Form.CREATE_BACKUPS ).setEnabled( false );
   		dialog.getFormField( Form.RECREATE_OPTIONAL ).setEnabled( false );
   		dialog.getFormField( Form.RECREATE_REQUESTS ).addFormFieldListener( new XFormFieldListener() {

				public void valueChanged( XFormField sourceField, String newValue, String oldValue )
				{
					boolean enabled = dialog.getBooleanValue( Form.RECREATE_REQUESTS );
					
					dialog.getFormField( Form.CREATE_BACKUPS ).setEnabled( enabled );
					dialog.getFormField( Form.RECREATE_OPTIONAL ).setEnabled( enabled );
				}} );
   	}
   	
   	dialog.setValue( Form.DEFINITION_URL, iface.getDefinition() );
   	if( !dialog.show() )
   		return;
   	
      String url = dialog.getValue( Form.DEFINITION_URL );
      if (url == null || url.trim().length() == 0 )
         return;

      try
		{
			File file = new File( url );
			if( file.exists() ) 
				url = file.toURL().toString();
		}
		catch( Exception e )
		{
			SoapUI.logError( e );
		}
      	
      boolean createRequests = dialog.getBooleanValue( Form.CREATE_REQUESTS );
      
      try
		{
      	if( iface.updateDefinition( url, createRequests ))
      	{
      		if( dialog.getBooleanValue( Form.RECREATE_REQUESTS ))
      		{
	      		int cnt = recreateRequests( iface, dialog.getBooleanValue( Form.RECREATE_OPTIONAL ),
	      					dialog.getBooleanValue( Form.CREATE_BACKUPS )	);
	      		
	      		UISupport.showInfoMessage(	"Update of interface successfull, [" + cnt + "] requests have" +
	      				" been udpated.",  "Update Definition" );
      		}
      		else
      		{
      			UISupport.showInfoMessage(	"Update of interface successfull",  "Update Definition" );
      		}
      	}
      	else
      		UISupport.showInfoMessage(	"Update of interface failed",  "Update Definition" );
		}
		catch (Exception e1)
		{
			UISupport.showInfoMessage(	"Failed to update interface: [" + e1 + "]", "Update Definition" );
			SoapUI.logError( e1 );
		}
   }

	protected int recreateRequests( WsdlInterface iface, boolean buildOptional, boolean createBackups )
	{
		int count = 0;
		
		// first check operations
		for( int c = 0; c < iface.getOperationCount(); c++ )
		{
			WsdlOperation operation = iface.getOperationAt( c );
			String newRequest = operation.createRequest( buildOptional );
			List<Request> requests = operation.getRequests();
			
			for( Request request : requests )
			{
				String requestContent = request.getRequestContent();
				String req = XmlUtils.transferValues( requestContent, newRequest );
				
				// changed?
				if( !req.equals( requestContent ) )
				{
					if( !XmlUtils.prettyPrintXml( req ).equals( XmlUtils.prettyPrintXml( requestContent )))
					{
						if( createBackups )
						{
							WsdlRequest backupRequest = operation.addNewRequest( 
										"Backup of [" + request.getName() + "]" );
							((WsdlRequest)request).copyTo( backupRequest, false, false );
						}
						
						((WsdlRequest)request).setRequestContent( req );
						count++;
					}
				}
			}
		}
		
		// now check testsuites..
		for( TestSuite testSuite : iface.getProject().getTestSuites() )
		{
			for( TestCase testCase : testSuite.getTestCaseList() )
			{
				for( int c = 0; c < testCase.getTestStepCount(); c++ )
				{
					WsdlTestStep testStep = ( WsdlTestStep ) testCase.getTestStepAt( c );
					if( testStep instanceof WsdlTestRequestStep )
					{
						WsdlTestRequest testRequest = ((WsdlTestRequestStep)testStep).getTestRequest();
						String newRequest = testRequest.getOperation().createRequest( buildOptional );
						
						String req = XmlUtils.transferValues( testRequest.getRequestContent(), newRequest );
						
						// changed?
						if( !req.equals( testRequest.getRequestContent() ) )
						{
							if( createBackups )
							{
								((WsdlTestCase)testCase).importTestStep( testStep,
											"Backup of [" + testStep.getName() + "]", -1 ).setDisabled( true );
							}
							
							((WsdlRequest)testRequest).setRequestContent( req );
							count++;
						}
					}
				}
			}
		}
		
		return count;
	} 
   
   @AForm(description = "Specify Update Definition options", name = "Update Definition" )
	protected interface Form
	{
   	@AField( name="Definition URL", description = "The URL or file for the updated definition", type=AFieldType.FILE )
		public final static String DEFINITION_URL = "Definition URL";
   	
   	@AField( name="Create New Requests", description = "Create default requests for new methods", type=AFieldType.BOOLEAN )
		public final static String CREATE_REQUESTS = "Create New Requests";
   	
   	@AField( name="Recreate Requests", description = "Recreate existing request with the new schema", type=AFieldType.BOOLEAN )
		public final static String RECREATE_REQUESTS = "Recreate Requests";
   	
   	@AField( name="Recreate Optional", description = "Recreate optional content when updating requests", type=AFieldType.BOOLEAN )
		public final static String RECREATE_OPTIONAL = "Recreate Optional";
   	
   	@AField( name="Create Backups", description = "Create backup copies of changed requests", 
   				 type=AFieldType.BOOLEAN )
		public final static String CREATE_BACKUPS = "Create Backups";
	}
}
