/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.mock;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.util.Enumeration;

import javax.mail.MessagingException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.xmlbeans.XmlCursor;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;

import com.eviware.soapui.SoapUI;
import com.eviware.soapui.impl.wsdl.WsdlOperation;
import com.eviware.soapui.impl.wsdl.submit.transports.http.MockRequestDataSource;
import com.eviware.soapui.impl.wsdl.submit.transports.http.MultipartMessageSupport;
import com.eviware.soapui.impl.wsdl.support.soap.SoapVersion;
import com.eviware.soapui.model.iface.Attachment;
import com.eviware.soapui.model.mock.MockRequest;
import com.eviware.soapui.support.StringUtils;
import com.eviware.soapui.support.types.StringToStringMap;

/**
 * Request-class created when receiving an external request to a WsdlMockService
 * 
 * @author ole.matzura
 */

public class WsdlMockRequest implements MockRequest
{
	private StringToStringMap requestHeaders;
	private String requestContent;
	private MultipartMessageSupport mmSupport;
	private XmlObject requestXmlObject;
	private SoapVersion soapVersion;
	private final HttpServletResponse response;
	private String protocol;
	private String path;
	private final WsdlMockRunContext context;
	private final WsdlMockRunContext requestContext;
	private final HttpServletRequest request;

	public WsdlMockRequest( HttpServletRequest request, HttpServletResponse response, WsdlMockRunContext context ) throws Exception
	{
		this.request = request;
		this.response = response;
		this.context = context;
		
		requestContext = new WsdlMockRunContext( context.getMockService(), null );
		
		requestHeaders = new StringToStringMap();
		for( Enumeration e = request.getHeaderNames(); e.hasMoreElements(); )
		{
			String header = ( String ) e.nextElement();
			requestHeaders.put( header, request.getHeader( header ) );
		}
		
		protocol = request.getProtocol();
		path = request.getPathInfo();
		
		String contentType = request.getContentType();
		
		if( contentType != null &&	contentType.toUpperCase().startsWith( "MULTIPART" ))
		{
			readMultipartRequest( request );
			contentType = mmSupport.getRootPart().getContentType();
		}
		else
		{
			this.requestContent = readRequestContent( request );
		}
		
		initSoapVersion( contentType );
	}
	
	/**
	 * Init soapversion from content-type header.. should envelope be checked and/or override? 
	 */
	
	protected boolean initSoapVersion( String contentType )
	{
		soapVersion = contentType.startsWith( SoapVersion.Soap11.getContentType() ) ? SoapVersion.Soap11 : null;
		soapVersion = soapVersion == null && contentType.startsWith( SoapVersion.Soap12.getContentType() ) ? SoapVersion.Soap12 : soapVersion;

		return true;
	}
	
	public SoapVersion getSoapVersion()
	{
		return soapVersion;
	}

	public String getProtocol()
	{
		return protocol;
	}

	private void readMultipartRequest( HttpServletRequest request ) throws MessagingException
	{
		StringToStringMap values = StringToStringMap.fromHttpHeader( request.getContentType() );
		mmSupport = new MultipartMessageSupport( new MockRequestDataSource( request ), values.get( "start" ), null, true );
	}

	private String readRequestContent( HttpServletRequest request ) throws Exception
	{
		String encoding = request.getCharacterEncoding();
		if( encoding == null )
			encoding = "UTF-8";
		else
			encoding = StringUtils.unquote( encoding );
		
		BufferedReader reader = new BufferedReader( new InputStreamReader( request.getInputStream(), encoding) );
		StringBuffer buf = new StringBuffer();
		String line = reader.readLine();
		while( line != null )
		{
			buf.append( line ).append( "\r\n" );
			line = reader.readLine();
		}
		
		return buf.toString();
	}


	public Attachment[] getRequestAttachments()
	{
		return mmSupport == null ? new Attachment[0] : mmSupport.getAttachments();
	}

	public String getRequestContent()
	{
		return mmSupport == null ? requestContent : mmSupport.getContentAsString();
	}

	public StringToStringMap getRequestHeaders()
	{
		return requestHeaders;
	}

	public void setRequestContent( String requestContent )
	{
		this.requestContent = requestContent;
	}

	public XmlObject getRequestXmlObject() throws XmlException
	{
		if( requestXmlObject == null )
			requestXmlObject = XmlObject.Factory.parse( getRequestContent() );
		
		return requestXmlObject;
	}

	public XmlObject getBodyElement() throws XmlException
	{
		XmlObject[] envelope = getRequestXmlObject().selectChildren( soapVersion.getEnvelopeQName() );
		if( envelope.length != 1 )
		    throw new XmlException( "Missing/Invalid SOAP Envelope, expecting [" + soapVersion.getEnvelopeQName() + "]" );
		
		XmlObject[] body = envelope[0].selectChildren( soapVersion.getBodyQName() );
		if( body.length != 1 )
		    throw new XmlException( "Missing/Invalid SOAP Body, expecting [" + soapVersion.getBodyQName() + "]" );
		
		return body[0];
	}

	public HttpServletResponse getHttpResponse()
	{
		return response;
	}

	public HttpServletRequest getHttpRequest() 
	{
		return request;
	}
	
	public XmlObject getContentElement() throws XmlException
	{
		XmlObject bodyElement = getBodyElement();
		if( bodyElement != null )
		{
			XmlCursor cursor = bodyElement.newCursor();
			
			try
			{
				if( cursor.toFirstChild() )
				{
					while( !cursor.isContainer() )
						cursor.toNextSibling();

					if( cursor.isContainer() )
					{
						return cursor.getObject();
					}
				}
			}
			catch( Exception e )
			{
				SoapUI.logError( e );
			}
			finally
			{
				cursor.dispose();
			}
		}
		
		return null;
	}

	public String getPath()
	{
		return path;
	}

	public WsdlMockRunContext getContext()
	{
		return context;
	}

	public void setOperation( WsdlOperation operation )
	{
		if( mmSupport != null )
			mmSupport.setOperation( operation );
	}

	public WsdlMockRunContext getRequestContext()
	{
		return requestContext;
	}
}
