/*
 *  soapUI, copyright (C) 2004-2007 eviware.com 
 *
 *  soapUI is free software; you can redistribute it and/or modify it under the 
 *  terms of version 2.1 of the GNU Lesser General Public License as published by 
 *  the Free Software Foundation.
 *
 *  soapUI is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
 *  even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 *  See the GNU Lesser General Public License for more details at gnu.org.
 */

package com.eviware.soapui.impl.wsdl.testcase;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.log4j.Logger;

import com.eviware.soapui.config.LoadTestConfig;
import com.eviware.soapui.config.TestCaseConfig;
import com.eviware.soapui.config.TestStepConfig;
import com.eviware.soapui.impl.wsdl.AbstractWsdlModelItem;
import com.eviware.soapui.impl.wsdl.WsdlTestSuite;
import com.eviware.soapui.impl.wsdl.loadtest.LoadTestAssertion;
import com.eviware.soapui.impl.wsdl.loadtest.WsdlLoadTest;
import com.eviware.soapui.impl.wsdl.loadtest.assertions.TestStepStatusAssertion;
import com.eviware.soapui.impl.wsdl.teststeps.WsdlTestStep;
import com.eviware.soapui.impl.wsdl.teststeps.registry.WsdlTestStepFactory;
import com.eviware.soapui.impl.wsdl.teststeps.registry.WsdlTestStepRegistry;
import com.eviware.soapui.model.support.PropertiesMap;
import com.eviware.soapui.model.testsuite.LoadTest;
import com.eviware.soapui.model.testsuite.TestCase;
import com.eviware.soapui.model.testsuite.TestRunListener;
import com.eviware.soapui.model.testsuite.TestStep;
import com.eviware.soapui.support.UISupport;
import com.eviware.soapui.support.action.swing.ActionList;
import com.eviware.soapui.support.action.swing.DefaultActionList;

/**
 * TestCase implementation for WSDL projects
 *  
 * @author Ole.Matzura
 */

public class WsdlTestCase extends AbstractWsdlModelItem<TestCaseConfig> implements TestCase
{
   private final static Logger logger = Logger.getLogger( WsdlTestCase.class );
   public final static String KEEP_SESSION_PROPERTY = WsdlTestCase.class.getName() + "@keepSession";
   public final static String FAIL_ON_ERROR_PROPERTY = WsdlTestCase.class.getName() + "@failOnError";
   public final static String FAIL_ON_ERRORS_PROPERTY = WsdlTestCase.class.getName() + "@failOnErrors";
   public final static String DISCARD_OK_RESULTS = WsdlTestCase.class.getName() + "@discardOkResults";
	private static final String SEARCH_PROPERTIES_PROPERTY = WsdlTestCase.class.getName() + "@searchProperties";
	
	private final WsdlTestSuite testSuite;
   private List<WsdlTestStep> testSteps = new ArrayList<WsdlTestStep>();
   private List<WsdlLoadTest> loadTests = new ArrayList<WsdlLoadTest>();
   private Set<TestRunListener> testRunListeners = new HashSet<TestRunListener>();
	private DefaultActionList createActions;
	private final boolean forLoadTest;
	
   public WsdlTestCase(WsdlTestSuite testSuite, TestCaseConfig config, boolean forLoadTest )
   {
   	super( config, testSuite, "/testCase.gif" );
   	
      this.testSuite = testSuite;
		this.forLoadTest = forLoadTest;
      
      List<TestStepConfig> testStepConfigs = config.getTestStepList();
      for (TestStepConfig tsc : testStepConfigs )
      {
         WsdlTestStep testStep = createTestStepFromConfig(tsc);
         if( testStep != null )
         {
            ensureUniqueName( testStep );
         	testSteps.add( testStep );
         }
      }
      
      for( TestStep step : testSteps )
      {
      	WsdlTestStep testStep = (WsdlTestStep) step;
      	testStep.postInit( testStep.getConfig() );
      }
      
      if( !forLoadTest )
      {
	      List<LoadTestConfig> loadTestConfigs = config.getLoadTestList();
	      for (LoadTestConfig tsc : loadTestConfigs)
	      {
	         WsdlLoadTest loadTest = new WsdlLoadTest( this, tsc );
				loadTests.add( loadTest);
	      }
	
//	      addAction( new ShowDesktopPanelAction( "Open TestCase Editor", 
//	      		"Opens the TestCase Editor for this TestCase", this ));
//	      addAction( ActionSupport.SEPARATOR_ACTION );
//	      addAction( new AddNewLoadTestAction( this ));
//	      addAction( new CloneTestCaseAction( this ));
//	      addAction( new ClearTestCaseAction( this ));
//	      addAction( ActionSupport.SEPARATOR_ACTION );
//	      addAction( new TestCaseOptionsAction( this ));
	
//	      createActions = new DefaultActionList( "New Step" );
//			WsdlTestStepRegistry registry = WsdlTestStepRegistry.getInstance();
//			WsdlTestStepFactory[] factories = (WsdlTestStepFactory[]) registry.getFactories();
//	
//			for (int c = 0; c < factories.length; c++)
//			{
//				if (factories[c].canCreate())
//					createActions.addAction( new AddTestStepAction(factories[c]));
//			}
//	
//			addAction( new ActionSupport.ActionListAction( createActions ));
//	      addAction( ActionSupport.SEPARATOR_ACTION );
//	      addAction( new RenameTestCaseAction( this ) );
//	      addAction( new DeleteTestCaseAction( this ) );
//	      addAction( ActionSupport.SEPARATOR_ACTION );
//	      addAction( new ShowOnlineHelpAction( HelpUrls.TESTCASE_HELP_URL ));
      }
      
      // init default configs
      if( !config.isSetFailOnError() )
      	config.setFailOnError( true );
      
      if( !config.isSetFailTestCaseOnErrors() )
      	config.setFailTestCaseOnErrors( true );

      if( !config.isSetKeepSession() )
      	config.setKeepSession( false );
   }

   public boolean getKeepSession()
   {
   	return getConfig().getKeepSession();
   }
   
   public void setKeepSession( boolean keepSession )
   {
   	boolean old = getKeepSession();
   	if( old != keepSession )
   	{
   		getConfig().setKeepSession( keepSession );
   		notifyPropertyChanged( KEEP_SESSION_PROPERTY, old, keepSession );
   	}
   }
   
   public boolean getFailOnError()
   {
   	return getConfig().getFailOnError();
   }
   
   public boolean getFailTestCaseOnErrors()
   {
   	return getConfig().getFailTestCaseOnErrors();
   }
   
   public void setFailOnError( boolean failOnError )
   {
   	boolean old = getFailOnError();
   	if( old != failOnError )
   	{
   		getConfig().setFailOnError( failOnError );
   		notifyPropertyChanged( FAIL_ON_ERROR_PROPERTY, old, failOnError );
   	}
   }
   
   public void setFailTestCaseOnErrors( boolean failTestCaseOnErrors )
   {
   	boolean old = getFailTestCaseOnErrors();
   	if( old != failTestCaseOnErrors )
   	{
   		getConfig().setFailTestCaseOnErrors( failTestCaseOnErrors );
   		notifyPropertyChanged( FAIL_ON_ERRORS_PROPERTY, old, failTestCaseOnErrors );
   	}
   }
   
   public boolean getSearchProperties()
   {
   	return getConfig().getSearchProperties();
   }
   
   public void setSearchProperties( boolean searchProperties )
   {
   	boolean old = getSearchProperties();
   	if( old != searchProperties )
   	{
   		getConfig().setSearchProperties( searchProperties );
   		notifyPropertyChanged( SEARCH_PROPERTIES_PROPERTY, old, searchProperties );
   	}
   }
   
   public boolean getDiscardOkResults()
   {
   	return getConfig().getDiscardOkResults();
   }
   
   public void setDiscardOkResults( boolean discardOkResults )
   {
   	boolean old = getDiscardOkResults();
   	if( old != discardOkResults )
   	{
   		getConfig().setDiscardOkResults( discardOkResults );
   		notifyPropertyChanged( DISCARD_OK_RESULTS, old, discardOkResults );
   	}
   }
   
	private WsdlTestStep createTestStepFromConfig(TestStepConfig tsc )
	{
		WsdlTestStepFactory factory = WsdlTestStepRegistry.getInstance().getFactory( tsc.getType() );
		if( factory != null )
		{
			WsdlTestStep testStep = factory.buildTestStep( this, tsc, forLoadTest );
			return testStep;
		}
		else 
		{
			logger.error( "Failed to create test step for [" + tsc.getName() + "]" );
			return null;
		}
	}

	private boolean ensureUniqueName(WsdlTestStep testStep)
	{
		String name = testStep.getName();
		while( name == null || getTestStepByName( name ) != null )
		{
			if( name == null )
				name = testStep.getName();
			else
			{
				int cnt = 0;
				
				while( getTestStepByName( name ) != null )
				{
					cnt++;
					name = testStep.getName() + " " + cnt;
				}
				
				if( cnt == 0 )
					break;
			}
			
			name = UISupport.prompt( "TestStep name must be unique, please specify new name for step\n" +
					"[" + testStep.getName() + "] in TestCase [" + getTestSuite().getProject().getName() + "->" +
					getTestSuite().getName() + "->" + getName() + "]", 
					"Change TestStep name", name);
			
			if( name == null )
				return false;
		}
		
		if( !name.equals( testStep.getName() ))
			testStep.setName( name );
		
		return true;
	}

   public WsdlLoadTest addNewLoadTest( String name )
   {
      WsdlLoadTest loadTest = new WsdlLoadTest( this, getConfig().addNewLoadTest());
      loadTest.setStartDelay( 0 );
      loadTest.setName( name );
      loadTests.add( loadTest );
      
      loadTest.addAssertion( TestStepStatusAssertion.STEP_STATUS_TYPE, 
      		LoadTestAssertion.ANY_TEST_STEP, false );

      ((WsdlTestSuite)getTestSuite()).fireLoadTestAdded( loadTest );
      
      return loadTest;
   }
   
	public void removeLoadTest(WsdlLoadTest loadTest)
	{
		int ix = loadTests.indexOf( loadTest );

		loadTests.remove( ix );
		
		try
		{
			((WsdlTestSuite)getTestSuite()).fireLoadTestRemoved( loadTest );
		}
		finally
		{
			loadTest.release();
			getConfig().removeLoadTest( ix );
		}
	} 

   public WsdlTestSuite getTestSuite()
   {
      return testSuite;
   }

   public WsdlTestStep cloneStep( WsdlTestStep testStep, String name )
   {
   	return testStep.clone( this, name );
   }
   
   public WsdlTestStep getTestStepAt(int index)
   {
      return testSteps.get( index );
   }

   public int getTestStepCount()
   {
      return testSteps.size();
   }

   public WsdlLoadTest getLoadTestAt(int index)
   {
      return loadTests.get( index );
   }

   public LoadTest getLoadTestByName(String loadTestName)
	{
		return (LoadTest) getWsdlModelItemByName( loadTests, loadTestName );
	}

	public int getLoadTestCount()
   {
      return loadTests.size();
   }
   
   public WsdlTestStep addTestStep( TestStepConfig stepConfig )
   {
   	return insertTestStep( stepConfig, -1 );
   }
   
   public WsdlTestStep addTestStep( String type, String name  )
   {
   	TestStepConfig newStepConfig = WsdlTestStepRegistry.getInstance().getFactory( type ).createNewTestStep( this, name );
   	if( newStepConfig != null ) 
   	{
   		return addTestStep( newStepConfig);
  		}
   	else return null;
   }
   
   public WsdlTestStep insertTestStep( String type, String name, int index  )
   {
   	TestStepConfig newStepConfig = WsdlTestStepRegistry.getInstance().getFactory( type ).createNewTestStep( this, name );
   	if( newStepConfig != null )
   	{
   		return insertTestStep( newStepConfig, index );
   	}
   	else return null;
   }
   
   public WsdlTestStep importTestStep( WsdlTestStep testStep, String name, int index  )
   {
   	testStep.onSave();
   	TestStepConfig newStepConfig = ( TestStepConfig ) testStep.getConfig().copy();
		newStepConfig.setName( name );
		return insertTestStep( newStepConfig, index );
   }
   
   public WsdlTestStep [] importTestSteps( WsdlTestStep [] testSteps, int index  )
   {
   	TestStepConfig [] newStepConfigs = new TestStepConfig[testSteps.length];
   	
   	for( int c = 0; c < testSteps.length; c++ )
   	{
   		testSteps[c].onSave();
   		newStepConfigs[c] = ( TestStepConfig ) testSteps[c].getConfig().copy();
   	}
   	
   	return insertTestSteps( newStepConfigs, index );
   }
   
   public WsdlTestStep insertTestStep( TestStepConfig stepConfig, int ix )
   {
   	TestStepConfig newStepConfig = ix == -1 ? getConfig().addNewTestStep() : getConfig().insertNewTestStep( ix );
   	newStepConfig.set( stepConfig );
   	WsdlTestStep testStep = createTestStepFromConfig( newStepConfig );
   	
   	if( !ensureUniqueName( testStep ))
   		return null;
   	
      if( ix == -1 )
      	testSteps.add( testStep );
      else
      	testSteps.add( ix, testStep );
      
      testStep.postInit( testStep.getConfig() );
      
      if( getTestSuite() != null )
      	((WsdlTestSuite)getTestSuite()).fireTestStepAdded( testStep, ix == -1 ? testSteps.size()-1 : ix  );
      
      return testStep;
   }
   
   public WsdlTestStep [] insertTestSteps( TestStepConfig [] stepConfig, int ix )
   {
   	WsdlTestStep [] result = new WsdlTestStep[stepConfig.length];
   	
   	for( int c = 0; c < stepConfig.length; c++ )
   	{
   		TestStepConfig newStepConfig = ix == -1 ? getConfig().addNewTestStep() : getConfig().insertNewTestStep( ix+c );
   		newStepConfig.set( stepConfig[c] );
   		WsdlTestStep testStep = createTestStepFromConfig( newStepConfig );
   	
	   	if( !ensureUniqueName( testStep ))
	   		return null;
	   	
	      if( ix == -1 )
	      	testSteps.add( testStep );
	      else
	      	testSteps.add( ix+c, testStep );
	      
	      result[c] = testStep;
   	}
      
   	for( int c = 0; c < result.length; c++ )
   	{
   		result[c].postInit( result[c].getConfig() );
	      
	      if( getTestSuite() != null )
	      	((WsdlTestSuite)getTestSuite()).fireTestStepAdded( result[c], getIndexOfTestStep( result[c] ));
   	}
      
      return result;
   }
   
   public void removeTestStep(WsdlTestStep testStep)
   {
      int ix = testSteps.indexOf( testStep );
      if( ix == -1 )
      {
         logger.error( "TestStep [" + testStep.getName() + "] passed to removeTestStep in testCase [" + 
               getName() + "] not found" );
         return;
      }
      
      testSteps.remove( ix );

      try
      {
      	((WsdlTestSuite)getTestSuite()).fireTestStepRemoved( testStep, ix );
      }
      finally
      {
      	testStep.release();
      
	      for( int c = 0; c < getConfig().sizeOfTestStepArray(); c++ ) 
	      {
	      	if( testStep.getConfig() == getConfig().getTestStepArray( c ))
	      	{
	      		getConfig().removeTestStep( c );
	      		break;
	      	}
	      }
      }
   }
   
   public WsdlTestCaseRunner run( PropertiesMap properties, boolean async ) 
   {
   	WsdlTestCaseRunner runner = new WsdlTestCaseRunner( this, properties );
   	runner.start( async );
   	return runner;
	}

	public void addTestRunListener(TestRunListener listener) {
		if( listener == null )
			throw new RuntimeException( "listener must not be null" );
		
		testRunListeners.add( listener );
	}

	public void removeTestRunListener(TestRunListener listener) {
		testRunListeners.remove( listener );
	}
	
	public TestRunListener [] getTestRunListeners()
	{
		return testRunListeners.toArray( new TestRunListener[testRunListeners.size()]);
	}

	WsdlTestStep[] getTestSteps() 
	{
		return testSteps.toArray( new WsdlTestStep[testSteps.size()] );
 	}

	public int getIndexOfTestStep(TestStep step)
	{
		return testSteps.indexOf( step );
	}
	
	/**
	 * Moves a step by the specified offset, a bit awkward since xmlbeans doesn't support reordering
	 * of arrays, we need to create copies of the contained XmlObjects
	 * 
	 * @param ix
	 * @param offset
	 */
	
	public void moveTestStep(int ix, int offset)
	{
		if( offset == 0 ) return;
		WsdlTestStep step = (WsdlTestStep) testSteps.get( ix );

		if( ix + offset >= testSteps.size() )
			offset = testSteps.size()-ix-1;
		
		testSteps.remove( ix );
		testSteps.add( ix+offset, step );

		TestStepConfig [] configs = new TestStepConfig[testSteps.size()];
		
		for( int c = 0; c < testSteps.size(); c++ )
		{
			if( offset > 0 )
			{
				if( c < ix )
					configs[c] = (TestStepConfig) getConfig().getTestStepArray(c).copy();
				else if( c < (ix+offset))
					configs[c] = (TestStepConfig) getConfig().getTestStepArray(c+1).copy();
				else if( c == ix+offset )
					configs[c] = (TestStepConfig) getConfig().getTestStepArray(ix).copy();
				else
					configs[c] = (TestStepConfig) getConfig().getTestStepArray(c).copy();
			}
			else
			{
				if( c < ix+offset )
					configs[c] = (TestStepConfig) getConfig().getTestStepArray(c).copy();
				else if( c == ix+offset )
					configs[c] = (TestStepConfig) getConfig().getTestStepArray(ix).copy();
				else if( c <= ix )
					configs[c] = (TestStepConfig) getConfig().getTestStepArray(c-1).copy();
				else
					configs[c] = (TestStepConfig) getConfig().getTestStepArray(c).copy();				
			}
		}
		
		getConfig().setTestStepArray( configs );
		for( int c = 0; c < configs.length; c++ )
		{
			((WsdlTestStep) testSteps.get( c )).resetConfigOnMove( getConfig().getTestStepArray( c ));
		}
		
		((WsdlTestSuite)getTestSuite()).fireTestStepMoved(step, ix, offset );
	}
	
	public int getIndexOfLoadTest( LoadTest loadTest )
	{
		return loadTests.indexOf( loadTest );
	}


	public int getTestStepIndexByName(String stepName)
	{
		for( int c = 0; c < testSteps.size(); c++ )
		{
			if( testSteps.get( c ).getName().equals( stepName ))
				return c;
		}
		
		return -1;
	}
	
	public TestStep findPreviousStepOfType( TestStep referenceStep, Class stepClass )
	{
		int currentStepIndex = getIndexOfTestStep( referenceStep );
		int ix = currentStepIndex - 1 ;
		while( ix >= 0 && !getTestStepAt( ix ).getClass().equals( stepClass ))
		{
			ix--;
		}
		
		return ix < 0 ? null : getTestStepAt( ix );
	}
	
	public TestStep findNextStepOfType( TestStep referenceStep, Class stepClass )
	{
		int currentStepIndex = getIndexOfTestStep( referenceStep );
		int ix = currentStepIndex + 1 ;
		while( ix < getTestStepCount() && !getTestStepAt( ix ).getClass().equals( stepClass ))
		{
			ix++;
		}
		
		return ix >= getTestStepCount() ? null : getTestStepAt( ix );
	}

	public List<TestStep> getTestStepList()
	{
		List<TestStep> result = new ArrayList<TestStep>();
		for( TestStep step : testSteps )
			result.add( step );
			
		return result;
	}
	
	public List<TestStep> getTestStepsOfType( Class<? extends WsdlTestStep> stepType )
	{
		List<TestStep> result = new ArrayList<TestStep>();
		for( TestStep step : testSteps )
			if( step.getClass().isAssignableFrom( stepType ) )
				result.add( step );
			
		return result;
	}

	public WsdlTestStep getTestStepByName(String stepName)
	{
		return (WsdlTestStep)getWsdlModelItemByName( testSteps, stepName );
	}

	public WsdlLoadTest cloneLoadTest(WsdlLoadTest loadTest, String name)
	{
		loadTest.onSave();
		
		LoadTestConfig loadTestConfig = getConfig().addNewLoadTest();
		loadTestConfig.set( loadTest.getConfig().copy());
		
		WsdlLoadTest newLoadTest = new WsdlLoadTest( this, loadTestConfig );
		newLoadTest.setName( name );
	   loadTests.add( newLoadTest );

	   ((WsdlTestSuite)getTestSuite()).fireLoadTestAdded( newLoadTest );
	      
	   return newLoadTest;
	}

	public void release()
	{
		super.release();
		
		for( WsdlTestStep testStep : testSteps )
			testStep.release();
		
		for( WsdlLoadTest loadTest : loadTests )
			loadTest.release();
	}
	
	public ActionList getCreateActions()
	{
		return createActions;
	}

	public void resetConfigOnMove( TestCaseConfig testCaseArray )
	{
		setConfig( testCaseArray );
		int mod = 0;
		
		List<TestStepConfig> configs = getConfig().getTestStepList();
		for( int c = 0; c < configs.size(); c++ )
		{
			if( WsdlTestStepRegistry.getInstance().hasFactory( configs.get( c ) ))
			{
				((WsdlTestStep) testSteps.get( c-mod )).resetConfigOnMove( configs.get( c ));
			}
			else mod++;
		}
		
		List<LoadTestConfig> loadTestConfigs = getConfig().getLoadTestList();
		for( int c = 0; c < loadTestConfigs.size(); c++ )
		{
			loadTests.get( c ).resetConfigOnMove( loadTestConfigs.get( c ));
		}
	}

	@Override
	public void onSave()
	{
		for( WsdlTestStep testStep : testSteps )
			testStep.onSave();
		
		for( WsdlLoadTest loadTest : loadTests )
			loadTest.onSave();
	}
}
